package test.guiexp;


import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Properties;
import java.util.Random;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.joda.time.DateTime;
import org.qcg.broker.schemas.schedulingplan.types.AllocationStatus;

import schedframe.resources.ResourceDescription;
import schedframe.resources.ResourceStateDescription;
import schedframe.resources.units.Processors;
import schedframe.resources.units.ResourceUnit;
import schedframe.scheduling.JobInterface;
import schedframe.scheduling.Offer;
import schedframe.scheduling.Queue;
import schedframe.scheduling.Reservation;
import schedframe.scheduling.ResourceRequirements;
import schedframe.scheduling.TaskInterface;
import schedframe.scheduling.TimeRequirements;
import schedframe.scheduling.TimeResourceAllocation;
import schedframe.scheduling.events.SchedulingEvent;
import schedframe.scheduling.events.TaskCanceledEvent;
import schedframe.scheduling.plan.SchedulingPlanInterface;
import schedframe.scheduling.plan.impl.Allocation;
import schedframe.scheduling.plan.impl.Host;
import schedframe.scheduling.plan.impl.ScheduledTask;
import schedframe.scheduling.plan.impl.SchedulingPlan;
import schedframe.scheduling.plugin.SchedulingPluginConfiguration;
import schedframe.scheduling.plugin.configuration.DefaultConfiguration;
import schedframe.scheduling.plugin.grid.GridSchedulingPlugin;
import schedframe.scheduling.plugin.grid.JobRegistry;
import schedframe.scheduling.plugin.grid.Module;
import schedframe.scheduling.plugin.grid.ModuleList;
import schedframe.scheduling.plugin.grid.NetworkManager;
import schedframe.scheduling.plugin.grid.Prediction;
import schedframe.scheduling.plugin.grid.ReservationManager;
import schedframe.scheduling.plugin.grid.ResourceDiscovery;
import schedframe.scheduling.utils.ResourceParameterName;
import simulator.GridSchedulingSimulator;

public class GridARResourcePriorityPlugin implements GridSchedulingPlugin<org.qcg.broker.schemas.schedulingplan.SchedulingPlan>{

	private static Log log = LogFactory.getLog(GridARResourcePriorityPlugin.class);
	
	//private static int TIMER = 3600;
	private static int START_TASK_ID = 0;
	private static int PERCENT_OF_AR_TASKS = 40;
	private static int RESERVATION_NO_INCREASE_STEP = 10;
	
	public static ArrayList<String> taskIDForReservations;
	public static HashMap<String, Long> task_endTime;
	
	private int x = 0;
	private int y = 0;


	protected List<ResourceDescription> availableResources;
	
	public GridARResourcePriorityPlugin(){

		//SchedulingPluginConfiguration spc = DefaultConfiguration.forGridPlugin();		
		//((GridConfig)spc).addServedEvent(SchedulingEventType.TIMER, new Integer(TIMER));
		
		PERCENT_OF_AR_TASKS = GridSchedulingSimulator.getSimulationRunNumber()*RESERVATION_NO_INCREASE_STEP;
		generateTaskIDForReservations(PERCENT_OF_AR_TASKS, START_TASK_ID);
		task_endTime = new HashMap<String, Long>();
	}
	public SchedulingPlanInterface<org.qcg.broker.schemas.schedulingplan.SchedulingPlan> schedule(
			SchedulingEvent event, 
			Queue<? extends JobInterface<?>> jobQueue,
			Queue<? extends TaskInterface<?>> taskQueue,
			JobRegistry jobRegistry, 
			ModuleList moduleList,
			Prediction prediction)
			throws Exception {
		
		SchedulingPlan plan = null;
		
		ResourceDiscovery resources = null;
		ReservationManager reservManager = null;
		for(int i = 0; i < moduleList.size(); i++){
			Module m = moduleList.get(i);
			switch(m.getType()){
				case RESERVATION_MANAGER: reservManager = (ReservationManager) m;
					break;
				case RESOURCE_DISCOVERY: resources = (ResourceDiscovery) m;
					break;
			}
		}
		
		// this method is called any time new event arrived. 
		// Events types which are expected to appear here are defined in plugin configuration.
		// See getConfiguration() method
		
		// choose correct method to serve the event
		switch(event.getType()){
			case TIMER: plan = scheduleNewTask(jobQueue, taskQueue, 
													jobRegistry, resources, 
													reservManager, null, 
													prediction); 
													break;
			case TASK_ARRIVED: plan = scheduleNewTask(jobQueue, taskQueue, 
													jobRegistry, resources, 
													reservManager, null, 
													prediction); 
				break;
			
			case TASK_CANCELED: TaskCanceledEvent e = (TaskCanceledEvent) event; 
								plan = rejectTask(e.getJobId(), e.getTaskId());
				break;
				
			default: log.info("Scheduling event " + event.getType().name() + " is not" +
					" supportd by the plugin ");
				break;
		}
		
		return plan;
		
	}

	
	protected SchedulingPlan scheduleNewTask(Queue<? extends JobInterface<?>> jobQueue,
			Queue<? extends TaskInterface<?>> taskQueue,
					JobRegistry jobRegistry, 
					ResourceDiscovery resources,
					ReservationManager reservManager,
					NetworkManager networkManager,
					Prediction prediction)
					throws Exception{

		// prepare plan with decision of allocating resources to tasks
		SchedulingPlan plan = new SchedulingPlan();
		int size = taskQueue.size();
		
		// iterate over task queue
		for(int i = 0; i < size; i++) {
			TaskInterface<?> task = taskQueue.remove(0);
			
			//determine if task should request a reservation
			if(withReservation(task.getId())){
				
				// prepare description of task time and resource requirements
				TimeRequirements timeRequirements = new TimeRequirements(task, x, y);
				ResourceRequirements resourceRequirements = new ResourceRequirements(task);
				task_endTime.put(task.getId(), timeRequirements.getEndMillis());
				
				// get offers from all local resources. Grid plugin MUST be aware of the format
				// and the way in which local resource (plugin) returns offers.
				List<Offer> offers = reservManager.getOffer(timeRequirements, resourceRequirements, null);
				boolean  timeFault = false;
				// choose the offer which provides enough free resources. Order of the list "offers" 
				// is not determined, therefore results my differ between executions. 
				Offer offer = null;
				for(int offerIdx = 0; offerIdx < offers.size() && offer == null; offerIdx++){
					offer = offers.get(offerIdx);
					
					//reservation start time is based on a submit time and wait time from swf(see TimeRequirements class for details)
					//check if a given start time make sense according to current time
					if(timeRequirements.getStartMillis() < new DateTime().getMillis())
					{
						offer = null;
						timeFault = true;
						break;
					}
					
					//accept offers from AR dedicated resource only
					if(!offer.getProvider().getProviderId().equals("compRes1"))
					{	
						offer = null;
						continue;
					}

					// local resource manager (plugin) returns all time slots between start and end time
					// declared in timeRequirements. Each time slot describes different resource usage.
					// Check if there is enough free resources between start and end time, and create
					// new offer - argument/request for create reservation method.
					offer = createReservationRequest(task, offer, timeRequirements, resourceRequirements);
				}
				
				//if AR dedicated resource doesn't meet task requirements, try second resource
				if(offer == null && !timeFault){
					for(int offerIdx = 0; offerIdx < offers.size() && offer == null; offerIdx++){
						offer = offers.get(offerIdx);
						if(offer.getProvider().getProviderId().equals("compRes2"))
							offer = createReservationRequest(task, offer, timeRequirements, resourceRequirements);
						else offer = null;
					}
				}
				
				// if none of the offers doesn't cover the initial time period then extend it
				if(offer == null){

					log.warn("Changing task time requirements: " + task.getJobId() + "_" + task.getId());
					//change task time requirements - extend the possible end period to 100 hours
					if(timeFault)
						timeRequirements = new TimeRequirements(task, -(new DateTime().getMillis()-timeRequirements.getStartMillis())/1000, 360000);
					else
						timeRequirements = new TimeRequirements(task, x, 360000);
					
					resourceRequirements = new ResourceRequirements(task);
					
					offers = reservManager.getOffer(timeRequirements, resourceRequirements, null);
					
					offer = null;
					
					//get offers from AR dedicated resource only
					for(int offerIdx = 0; offerIdx < offers.size() && offer == null; offerIdx++){
						offer = offers.get(offerIdx);
						if(!offer.getProvider().getProviderId().equals("compRes1"))
						{	
							offer = null;
							continue;
						}
						offer = createReservationRequest(task, offer, timeRequirements, resourceRequirements);
					}
					
					//skip task if there is no offer satisfying offer
					if(offer == null){
						log.error("skipping task: " + task.getJobId() + "_" + task.getId());
						continue;
					}
				}
				if(offer != null){
					// create reservation. Offer should have only these time resource allocations
					// which should be reserved - all unnecessary allocations must be removed. Allocations
					// may be also brand new objects, but remember to set correct provider info.
					// There is different reservation for different time allocation created.
					List<Reservation> reservations = reservManager.createReservation(offer, null);
					
					// get the first reservation. In this example, there is only one created
					Reservation r = reservations.get(0);
					
					// set information about task (job and task id) for which this reservation
					// was created
					r.setJobId(task.getJobId());
					r.setTaskId(task.getId());
					
					// commit reservation. Form now reservation can not expire.
					r = reservManager.commitReservation(r, null);
			
					// prepare allocation description in scheduling plan
					
					// information about destination host
					Host host = new Host();
					host.setHostname(r.getAllocatedResource().getProvider().getProviderId());
					
					// information about created reservation
					Allocation allocation = new Allocation();
					allocation.setProcessesCount(1);
					allocation.setReservation(r);
					
					allocation.setHost(host);
					
					// information about the task itself
					ScheduledTask scheduledTask = new ScheduledTask();
					scheduledTask.setTaskId(r.getTaskId());
					scheduledTask.setJobId(r.getJobId());
					scheduledTask.addAllocation(allocation);		
					
					plan.addTask(scheduledTask);
				}
			}
			else{
				//for nonAR tasks used load balancing strategy
				log.info("running without reservation task: " + task.getJobId() + "_" + task.getId());

				//get available resource. In this example, there are only two resources
				List<ResourceDescription> availableResources = resources.getResources(null);
				ResourceDescription rd1 = availableResources.get(0);
				ResourceDescription rd2 = availableResources.get(1);
				ResourceDescription rd;
				
				//obtain the least loaded resource
				if(rd1.getQueueLoad(null) > rd2.getQueueLoad(null))
					rd = rd2;
				else rd = rd1;

				Host host = new Host();
				host.setHostname(rd.getProvider().getProviderId());
				
				Allocation allocation = new Allocation();
				allocation.setProcessesCount(1);
				allocation.setHost(host);
				
				ScheduledTask scheduledTask = new ScheduledTask();
				scheduledTask.setTaskId(task.getId());
				scheduledTask.setJobId(task.getJobId());
				scheduledTask.addAllocation(allocation);		
				
				plan.addTask(scheduledTask);
			}

		}
		return plan;
	}
	
	protected SchedulingPlan rejectTask(String jobId, String taskId){
		// prepare plan with decision about task rejection.
		SchedulingPlan plan = new SchedulingPlan();
			// create task 
			ScheduledTask task = new ScheduledTask();
				task.setJobId(jobId);
				task.setTaskId(taskId);
				// set its status as rejected
				task.setStatus(AllocationStatus.REJECTED);
		plan.addTask(task);
		return plan;
	}
	
	
	private Offer createReservationRequest(TaskInterface<?> task, Offer offer, TimeRequirements timeRequirements, ResourceRequirements resourceRequirements){
		// This method assumes, that time allocations in offer argument covers continuous time interval.
		
		try {
			// determine time duration for which resource reservation will be created
			long reqDuration = (task.getLength()/(long)task.getCpuCntRequest())*1000;
			// this is helper variable. Its value will change during processing
			long duration = reqDuration;
			
			// get number of required processors
			int reqCpu = Double.valueOf(resourceRequirements.getCpuCntRequest()).intValue();
			DateTime start = null;
			
			// iterate through all time allocation.
			// This loop stops when long enough time interval is found
			for(int i = 0; i < offer.size() && duration > 0; i++){
				TimeResourceAllocation allocation = offer.get(i);
				ResourceUnit unit = allocation.getAllocatedResource().
												getResourceUnit(ResourceParameterName.CPUCOUNT);

				// check if free amount of the resource is enough to satisfy required amount
				if(unit.getFreeAmount() >= reqCpu){
					// if so, then shorten required duration, because this allocation satisfies resource requirements and
					// it can be "joined" with previous one in single, continuous time interval. 
					duration = duration - (allocation.getEndMillis() - allocation.getStartMillis());
					if(start == null)
						start = allocation.getStart();
				} else {
					// otherwise, continue searching from the beginning
					duration = reqDuration;
					start = null;
				}
			}

			// if duration is grater then 0, it means that there is no single and continuous time interval
			// which can be reserved the task
			if(duration > 0){
				log.error("There is not enough free resources between " + 
							timeRequirements.getStart() + " " + timeRequirements.getStart().getMillis() + 
							" and " + 
							timeRequirements.getEnd() + " " + timeRequirements.getEnd().getMillis() + 
							" in offer " + offer.getId());
				return null;
			}
			
			// create new offer, which will be used as an argument/request by reservation manager to create
			// reservation for task.
			int totalAmount = offer.get(0).getAllocatedResource().
											getResourceUnit(ResourceParameterName.CPUCOUNT).
											getAmount();
			
			ResourceStateDescription resDesc = new ResourceStateDescription(offer.getProvider());
			resDesc.addResourceUnit(new Processors(totalAmount, reqCpu, 1));
			
			TimeResourceAllocation allocation = new TimeResourceAllocation(start, start.plus(reqDuration));
			allocation.setAllocatedResource(resDesc);
			
			Offer retOffer = new Offer();
			retOffer.setProvider(offer.getProvider());
			retOffer.add(allocation);
			
			return retOffer;

		} catch (NoSuchFieldException e) {
			e.printStackTrace();
		}
		
		return null;
	}
	
	public SchedulingPluginConfiguration getConfiguration() {
		return DefaultConfiguration.forGridPlugin();
	}

	public String getPluginName() {
		return getClass().getName();
	}

	public void initPlugin(Properties properties) {
	}
	

	public void generateTaskIDForReservations(int probability, int startID){
		//if(taskIDForReservations == null)
		//{
			taskIDForReservations = new ArrayList<String>();
			Random rand = new Random(47);
			for(int i = startID; i<=startID+1000; i++){
				if(rand.nextInt(100) < probability)
					taskIDForReservations.add(Integer.toString(i));
			}
		//}
	}
	
	public boolean withReservation(String taskID){
		if(taskIDForReservations.contains(taskID))
			return true;
		return false;
	}
	
	public void setIntervals(int simulationID){
		x = simulationID%2 * 7200;
		y = simulationID%2 * 14400;
	}
	
}
